
/*
* Standard includes
* --------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/*
* My includes
* --------------------------------------------------------
*/

#include "stats_handler.h"
#include "random_variates.h"
#include "structures.c"
#include "listsort.c"

/*
**********************************************************
* 				PROTOTYPES
**********************************************************
*/

int count_descendants(struct node *node);
void read_infinite_sites(int sample_size, struct node *node, struct mutation_site **start_infinite_sites_list,struct mutation_site **end_infinite_sites_list);
void mark_the_carriers(struct node *node, int *carriers);
void free_infinite_sites(struct mutation_site **start_infinite_sites_list);
void report_infinite_sites(int sample_size, struct mutation_site *start_infinite_sites_list);
void print_tree(struct node *node);
int count_mutations(struct node *node, int sum);

void mark_infinite_alleles(struct node *root, int *highest_allele);
void count_infinite_alleles(struct node **tree, struct node *root, int genomes, int sample_size);

void allele_compactor(int *array, int *temp);
void quicksort_int(int array[], int left, int right);
void make_stat_space(struct statset **stats_array, int number_repeats);
void do_stats(int *population, int *temp, struct statset *stats_array, int repeats, int current_repeat);
void display_stats(struct statset *stats_array, int repeats);
int count_number_phenotypes(int *population);
int count_allele_appearance(int *population);
double count_shared_alleles(int *population);
double count_average_alleles(int *population);
double count_unshared_alleles(int *population);

/*
**********************************************************
* 				FUNCTIONS
**********************************************************
*/

/*
* make_stat_space()
* --------------------------------------------------------
* allocates memory for the stats
*/


void make_stat_space(struct statset **stats_array, int number_repeats){

(*stats_array)=(struct statset *)malloc(number_repeats*sizeof(struct statset));

}




/*
* count_descendants()
* --------------------------------------------------------
* recursive algorithm to count the number of descendants of a node in the sample
* doesn't currently have any use at all - it is a legacy of the original Hudson (1990) code
*/


int count_descendants(struct node *target_node){

	int sum=0;
   if(target_node->descendant1==NULL) return (1);
   sum+=count_descendants(target_node->descendant1);
   sum+=count_descendants(target_node->descendant2);
   return(sum);
}


/*
* read_infinite_sites()
* --------------------------------------------------------
* This needs some explanation :oP
* It is a recursive function that take a node and reads the mutations on the clade below.
* For each mutation it finds, it creates an entry in a linked list of mutations, and gives that mutation a position on the continuous line (0 -> 1).
* It then finds all the carriers of that mutation, and marks them.
* the result is stored as follows!

Some ASCII art!  read both -> ("carriers") and | ("next") as pointers. [][][] is an array, 0/1 indicates whether an individual carries the mutation, position is the location of that mutation

so, individual number 8 carries mutations at 0.123 and 0.345, but not 0.566 or 0.216


	(postion:0.123)->[0][0][0][0][1][1][1][1][0][0][0][0][0][0]...to sample_size
       |
   (postion:0.566)->[0][0][0][0][1][1][1][0][0][0][0][0][0][0]...to sample_size
       |
   (postion:0.345)->[0][0][1][1][0][0][0][1][1][0][0][0][0][0]...to sample_size
       |
	(postion:0.216)->[0][0][0][0][0][0][0][0][0][0][0][0][0][0]...to sample_size
       |
      ...
       |
       .
to total number of
mutations in the tree

*
*/



void read_infinite_sites(int sample_size, struct node *node, struct mutation_site **pointer_to_start ,struct mutation_site **pointer_to_end){

	int i,j;

   struct mutation_site *new_mutation;


   /*For each mutation on the branch above this node we need to create a new entry in the infinite_sites_list and mark all the carriers of that mutation in the final sample*/
   for (i=0;i<node->mutations;i++){

   //printf("\nFor %i of %i, found at node %1.5f",i+1,node->mutations,node->time);getchar();
   //printf("\nReading Inf Sites!");getchar();
   	/*malloc some space for the new list member*/
   	new_mutation=(struct mutation_site*)malloc(sizeof(struct mutation_site));
   	if (new_mutation==NULL){printf("Malloc of new_mutation (in infinite alleles failed!) (press <enter>)");getchar();exit(1);}
      new_mutation->next=NULL;
      new_mutation->carriers=NULL;


      /*malloc an array of <sample_size> to hold (yes/no) info of which sample members carry the mutation, and set contents of array to zero*/
      new_mutation->carriers=(int *)malloc(sample_size*sizeof(int));
      if (new_mutation->carriers==NULL){printf("Malloc of new_mutation->carriers (in infinite alleles) failed! (press <enter>)");getchar();exit(1);}
      for(j=0;j<sample_size;j++){
      	new_mutation->carriers[j]=0;
      }
      /*Record a position in the sequence for the mutation*/
   	new_mutation->location=random_0to1();

      /*... exactly what it says on the tin*/
      mark_the_carriers(node,new_mutation->carriers);

      /*SORT OUT THE LINKED LIST*/
      if((*pointer_to_end)==NULL){
         (*pointer_to_end)=new_mutation;
         (*pointer_to_start)=new_mutation;
         //printf("\nadding (1st) %i of %i, found at node %i, location %f\t",i+1,node->mutations,node->id, (*pointer_to_end)->location);
      }else{

         (*pointer_to_end)->next=new_mutation;
         (*pointer_to_end)=new_mutation;
          //printf("\nadding (later) %i of %i, found at node %i location %f\t",i+1,node->mutations,node->id,(*pointer_to_end)->location);
      }


   }

   /* Exit condition: if either descendant is null we're at the bottom*/
   if (node->descendant1==NULL)return;

   /* and head off down the tree*/
	read_infinite_sites(sample_size,node->descendant1,pointer_to_start,pointer_to_end);
   read_infinite_sites(sample_size,node->descendant2,pointer_to_start,pointer_to_end);

}






/*
* mark_the_carriers()
* --------------------------------------------------------
*  recursive algorithm that finds all the descendants of a node and marks them as "carriers" by setting the corresponding member of an int array to 1.
*  kind of biblical sounding function, eh?
*/

void mark_the_carriers(struct node *node, int *carriers){

	if (node->descendant1==NULL){
   	carriers[(node->id)]=1;
      return;
   }

   mark_the_carriers(node->descendant1,carriers);
   mark_the_carriers(node->descendant2,carriers);
}




/*
* free_infinite_sites()
* --------------------------------------------------------
* To free the memory from the hideous data-structure we created to hold the ifinite sites data.
*/


void free_infinite_sites(struct mutation_site **start_infinite_sites_list){

  struct mutation_site *kill;

  while(*start_infinite_sites_list){
       kill=*start_infinite_sites_list;
       (*start_infinite_sites_list)=(*start_infinite_sites_list)->next;
       free(kill->carriers);
       free(kill);
  }

}

/*
* report_infinite_sites()
* --------------------------------------------------------
* to print out all the individuals and all the mutations they carry
*/

void report_infinite_sites(int sample_size, struct mutation_site *start_infinite_sites_list){

     struct mutation_site *current_mutation;
     int mutation_counter, i;

     start_infinite_sites_list=listsort(start_infinite_sites_list,0);
   //printf("\nReporting Inf Sites!");getchar();

     //mutation_counter=0;
     /*
     current_mutation=start_infinite_sites_list;

     if(current_mutation==NULL){
         printf("\n<NONE>\t\t");
         for(i=0;i<sample_size;i++){
            printf("%i", 0);
         }
     }
     current_mutation=start_infinite_sites_list;
     while(current_mutation!=NULL){
         printf("\n%f\t", current_mutation->location);
         mutation_counter++;
         for(i=0;i<sample_size;i++){
            printf("%i", current_mutation->carriers[i]);
         }
     		current_mutation=current_mutation->next;
     }
     printf("\t%i\t", mutation_counter);
     */
     for (i=0;i<sample_size;i++){
     		printf("\n #%i\t",i);
     		current_mutation=start_infinite_sites_list;
     		while(current_mutation!=NULL){
         	printf("%i", current_mutation->carriers[i]);
         	current_mutation=current_mutation->next;
     		}
     }

}

/*
* print_tree()
* --------------------------------------------------------
* Primarily as a debugging tool
* recursive algorith to print the tree in a parenthetic format
*/

void print_tree(struct node *node){
   /*
   if ((node->descendant1!=NULL)&&(node->ancestor!=NULL)){
   	printf("%i  %i",(node->id),(node->allele));
      }
   */
	if (node->descendant1==NULL){
   	printf("(%i)",(node->allele));
      return;
   }

  	printf("(");

   print_tree(node->descendant1);
   print_tree(node->descendant2);

   printf(")");


}

/*
* count_mutations()
* --------------------------------------------------------
* Simple function to count all the mutations on the tree. In a inf alleles model this amounts to the number of segregating sites
*
*/

int count_mutations(struct node *node, int sum){
   if (node->descendant1==NULL) return (node->mutations);
   sum+=(node->mutations);
   sum+=count_mutations(node->descendant1,0);
   sum+=+count_mutations(node->descendant2,0);
	return sum;
}

/*
* mark_infinite_alleles()
* --------------------------------------------------------
* Simple function, marks all the terminal nodes with the allele that they are carrying.
*
*/


void mark_infinite_alleles(struct node *node, int *highest_allele){

		int i;

     if (node->ancestor==NULL){
   		node->allele=*highest_allele;
     }

     if ((node->mutations!=0)&&(node->ancestor!=NULL)){
     		for (i=0;i<node->mutations;i++)(*highest_allele)++;
   		node->allele=*highest_allele;
	  }

     if ((node->mutations==0)&&(node->ancestor!=NULL)){
     		node->allele=node->ancestor->allele;
	  }

     if (node->descendant1==NULL) return;


   mark_infinite_alleles(node->descendant1,highest_allele);
   mark_infinite_alleles(node->descendant2,highest_allele);

}

/*
* count_infinite_alleles()
* --------------------------------------------------------
* Simple function all the carriers of each allele.
*
*/


void count_infinite_alleles(struct node **tree, struct node *root, int genomes, int lineages_per_sample){

    int current_highest_allele,i,j;
    struct allele *alleles_list;

    alleles_list=(struct allele *)malloc(genomes*lineages_per_sample*sizeof(struct allele));
    for(i=0;i<genomes*lineages_per_sample;i++){alleles_list[i].allele=-1;alleles_list[i].occurances=0;}

    current_highest_allele=0;

    mark_infinite_alleles(root,&current_highest_allele);
    //for(i=0;i<genomes*sample_size;i++)printf("\nterminal node %i has allele %i",tree[i]->id,tree[i]->allele);getchar();
    /*for each terminal node*/
    for(i=0;i<genomes*lineages_per_sample;i++){
    	/*for each element in the alleles_list*/
    	for(j=0;j<genomes*lineages_per_sample;j++){
      	/*if we find a match then increment and break*/
    		if (alleles_list[j].allele==tree[i]->allele){
         	alleles_list[j].occurances++;
            //printf("\nallele is %i so incremented allele %i (now %i)",tree[i]->allele,alleles_list[j].allele,alleles_list[j].occurances);getchar();
            break;
         }
         /*if we've passed through the alleles already entered (ie the identifier is sentinel -1) and not found it, add it to the list, increment and break*/
         if (alleles_list[j].allele==-1){
         	alleles_list[j].allele=tree[i]->allele;
            alleles_list[j].occurances++;
            ///printf("\nallele is %i so added allele %i",tree[i]->allele,alleles_list[j].allele);getchar();
            break;}
		}

    }

    printf("\nalleles:\t");
    for(i=0;i<genomes*lineages_per_sample;i++){
    	if (alleles_list[i].allele!=-1)
      	printf("\t%i",alleles_list[i].allele);
    }
    printf("\nfrequencies:\t");
    for(i=0;i<genomes*lineages_per_sample;i++){if (alleles_list[i].allele!=-1)printf("\t%i",alleles_list[i].occurances);}

    free(alleles_list);
}


/*
* allele_compactor()
* --------------------------------------------------------
* takes an array of genotypes delimited with -1 and terminated by -2 and compacts it into allelic phenotypes
*  1342-11152-12244-13244-11133-1-2 becomes
*  1234-1125-124-1234-113-1-2
* denoting alleleic phenotypes: 234, 125, 124, 234, 13
*/

void allele_compactor(int *array, int *temp){     // takes an array in which individuals are separated by -1 and the population is terminated by -2

int x=0,left=0,right=0,i=0,y;                     // counters



    //routine to sort the alleles (numerically) within each individual
	while (array[x]!=-2){
		while (array[x++]!=-1){
   		right=x-1;
   	}
   	quicksort_int(array,left,right);
   	left=x;
	}

   x=0;
   y=0;

   do{                                                        // do while not at the end of the string
     if (array[x]!=array[(x+1)]){             // if this allele is not the same as the next we want to keep it in the temp string
        	temp[y]=array[x];                     // keeping it, and get ready for the nex member of the temp string
         y++;
     }
     x++;
     temp[x]=-2;                                                     //look at the next member of the incoming string
   } while(array[x]!=-2);                            //   at the end of the string


   i=0;                                       //copy temp into array, replacing it.
   while(temp[i]!=-2){
   	array[i]=temp[i++];
   }
   array[i]=-2;



  /*i=0;                                       //print
   printf("\n");
   while(array[i]!=-2){
   	if (array[i]==-1)printf("\n");
      else printf("%i ",array[i]);
      i++ ;
   }
   //getchar();


   /**/


}

//************************************************************************8
// A quicksort for sorting integers
//************************************************************************8


void quicksort_int(int array[], int left, int right)     //calls array array[], calls first member left and lsat right
	{
   int i,j;
   int x ;
   int temp;


   i=left; j=right;
   x=array[(left+right)/2];                             //makes char x contain middle array member
   do {
      while(	(array[i]<x) && (i<right)	) i++;          // while left member is smaller than middle members, and left memeber is smaller than right maximum keep looking right
      while(	(array[j]>x) && (j>left)	) j--;           // while right memeber is greater than middle members, and right member is greater than maximum left
      if (i<=j)
         {                                                         // if current left is less than or equal to current right
      	temp=array[i];                                          // put it in temp,
         array[i]=array[j];                                    // put right in left
         array[j]=temp;                                          // put temp in right
         i++;                                                      //
         j--;                                                      // move current left and right up
         }
      }
      while (i<=j);
      if (left<j) quicksort_int(array, left, j);
      if (i<right) quicksort_int(array, i, right);
}

/*
* do_stats()
* --------------------------------------------------------
* executes a collection of functions that operate on population[] to generate a set of stats that are stored in stats_array
*/

void do_stats(int *population, int *temp, struct statset *stats_array, int repeats, int current_repeat){

allele_compactor(population,temp);
stats_array[current_repeat].number_alleles_per_pop=count_allele_appearance(population);
stats_array[current_repeat].number_alleles_per_indiv=count_average_alleles(population);
stats_array[current_repeat].number_phenotypes=count_number_phenotypes(population);
stats_array[current_repeat].shared_alleles=count_shared_alleles(population);
stats_array[current_repeat].unshared_alleles=count_unshared_alleles(population);

//db
//printf("\nav alleles = %f and shared alleles= %f one by t'other %f", stats_array[current_repeat].number_alleles_per_indiv,stats_array[current_repeat].shared_alleles,stats_array[current_repeat].shared_alleles/stats_array[current_repeat].number_alleles_per_indiv);getchar();

}

/*
* display_stats()
* --------------------------------------------------------
* diplays stats stored in stats_array[]
*/
void display_stats(struct statset *stats_array, int repeats){

int i;

for (i=0;i<repeats;i++){
  //	printf("\n\t%i\t%f\t%i\t%f\t%f",stats_array[i].number_alleles_per_ pop, stats_array[i].number_alleles_per_indiv,stats_array[i].number_phenotypes,stats_array[i].shared_alleles,stats_array[i].unshared_alleles);
}

}


/*
* count_allele_appearance()
* --------------------------------------------------------
* counts the number of different alleles appearing in population
*/

int count_allele_appearance(int *population){

   int i,j, count=0;
   int alleles_seen[50]; // this is a risk - should we ever see more than 50 different alleles we'll have an overflow

   for(i=0;i<50;i++)alleles_seen[i]=-1;

   i=0;
	while (population[i]!=-2){

   	if (population[i]!=-1){

      	for(j=0;j<50;j++){
         	//printf("\n(i,j)=%i,%i pop=%i list=%i" ,i,j,population[i],alleles_seen[j]);getchar();
      		if (alleles_seen[j]==population[i]){
            	//printf("\n\t(i,j)=%i,%i pop=%i, list=%i, count=%i they match so breaking" ,i, j, population[i] ,alleles_seen[j],count);getchar();
            	break;
            }
            if (alleles_seen[j]==-1){
               //printf("\n\t(i,j)=%i,%i pop=%i, list=%i, count=%i nothing matched so save and increm" ,i, j, population[i] ,alleles_seen[j],count);getchar();
					alleles_seen[j]=population[i];
               count++;
               break;
            }
         }
		}
   	i++;
   }

   //printf("\ncount is %i\n\n",count);getchar();

	return count;
}

/*
* count_number_phenotypes
* --------------------------------------------------------
* counts the number of different allelic phenotypes in the population
*/

int count_number_phenotypes(int *population){

	int flag,i,j,x,left,right,p, phenotypes[100][MAX_ALLELES], count=0;  // overflow if there is more than 100 phenotypes


	for (i=0;i<100;i++)phenotypes[i][0]=-1; //zero the array


   i=0;
   j=0;
   x=0;
   left=0;
   right=0;
   p=0;
   flag=0;

   while (population[i]!=-2){
      flag=1;
   	left=i;
      while(population[i]!=-1)i++;
      right=i++;
      j=0;
      //printf("\nfound left and right as %i and %i",left,right);getchar();
      while(flag){
         if(phenotypes[j][0]==-1){
            //printf("\nphenotypes[%i][0] was %i",j,phenotypes[j][0]);
            for(x=left;x<(right+1);x++)phenotypes[j][(x-left)]=population[x];
            //printf("\nsaved ");for(x=left;x<(right+1);x++){printf("%i ",phenotypes[j][(x-left)]);}
            break;
         }else{
            //printf("\nphenotypes[%i][0] was %i, ",j,phenotypes[j][0]);
            flag=0;
            for(x=left;x<(right+1);x++){
               //printf("\n Is phenotypes[%i][%i]=(%i) equal to population[%i]=(%i)?",j,(x-left),phenotypes[j][(x-left)],x,population[x]);
               if(phenotypes[j][(x-left)]!=population[x]){
                  //printf("\n\tphenotypes[j][%i]=(%i) != population[%i]=(%i)",(x-left),phenotypes[j][(x-left)],x,population[x]); getchar();
                	j++;
                  flag=1;
                  break;
               }
				}


         }
      }
	}

   count=0;
   while(phenotypes[count][0]!=-1)count++;

   //printf("\n\count is %i",count);getchar();

  return count;
}

/*
* count_shared_alleles()
* --------------------------------------------------------
* counts the average number of shared  alleles (in allelic phenotypes) for each pair of individuals in population[]
* remember population is in the form 0 1 2 -1 0 2 -1 1 2 3 -1 0 2 3 -1 0 2 -1 0 1 2 -1 -2
*
* where -1 terminates individuals and -2 terminates the end of the population.
*
* Does not count any pair twice, and it without replacement (ie does not campare individual n with its self)
*
*/


double count_shared_alleles(int *population){

   int count=0, individuals=0,starting_point=1,i,j,sentinel_counter=1;
   double average_per_indiv;
   i=0;
   while(population[i]!=-2){
   	sentinel_counter=1;
   	for(j=0;/*ever*/;j++){
         if (sentinel_counter>starting_point){
         	//printf("\nsent>start so in top loop");getchar();
         	if(population[j]==-1){j++;}
         	if(population[j]==-2)break;
				if(population[j]==population[i]){count++;}
         }
         if (sentinel_counter<=starting_point){
            //printf("\nsent<=start so in bottom loop");getchar();
            //printf(" i is %i, j is %i,",i,j);getchar();count++;
         	if(population[j]==-1)sentinel_counter++;
         }
      }
      i++;
      if(population[i]==-1){
      	i++;
		   individuals++;
         starting_point++;
      }
   }

   //printf("\nCount is %i, individuals %i and reurn value %f", count, individuals, (double)count/(0.5*((individuals*individuals)-individuals)));getchar();

   return (double)count/(0.5*((individuals*individuals)-individuals));
}

/*
* count_average_alleles()
* --------------------------------------------------------
* counts the average number of alleles in each allelic phenotype
*/

double count_average_alleles(int *population){

	int i=0, indiv=0, allele=0;

   while (population[i]!=-2){
   	if (population[i]==-1)indiv++;
      else allele++;
      i++;
   }

   //printf("\nresult= %f",(double)allele/indiv); getchar();
	return (double)allele/indiv;
}

/*
* count_unshared_alleles()
* --------------------------------------------------------
* counts the average number of unshared  alleles (in allelic phenotypes) for each pair of individuals in population[]
* remember population is in the form 0 1 2 -1 0 2 -1 1 2 3 -1 0 2 3 -1 0 2 -1 0 1 2 -1 -2
*
* where -1 terminates individuals and -2 terminates the end of the population.
*
* Does not count any pair twice, and is without replacement (ie does not compare individual n with its self)
*
* the method is far from clear :oP but there are two runthroughs, one from each point of view , and we only look at each pair once (from each POV)
*
*
*/


double count_unshared_alleles(int *population){

    int unshared=0,i=0,j,individuals=0,starting_point=1,sentinel_counter, shared_flag=0, ending_point=0;

    int blah[]={1,2,3,4,-1,1,-1,1,2,-1,3,4,-1,1,4,5,-1,-2}; //test array
    //population=blah; if we need to use a test array rather than a genuine one for debugging

    /*runthrough 1*/
    while(population[i]!=-2){                                     //until end of pop  (comparer)
   	sentinel_counter=1;                                         // counts individuals
   	for(j=0;/*ever*/;j++){                                      // until end of pop (comparand)
       if (sentinel_counter>starting_point){                      // if this individual is far enough to the right to be compared
         	if(population[j]==-1){
               if (shared_flag==0)unshared++;
               shared_flag=0;
            	j++;
				}
         	if(population[j]==-2)break;
            if(population[i]==population[j])shared_flag=1;
       }
         if (sentinel_counter<=starting_point){
            if(population[j]==-1){
            	shared_flag=0;
            	sentinel_counter++;
            }
         }
      }
      i++;
      if(population[i]==-1){
      	i++;
		   individuals++;
         starting_point++;
      }
   }

   i=0;j=0;sentinel_counter=0;ending_point=0;shared_flag=0;individuals=0;


   /*runthrough 2*/
     while(population[i]!=-2){                                     //until end of pop  (comparer)
   	sentinel_counter=0;                                         // counts individuals
   	for(j=0;/*ever*/;j++){                                      // until end of pop (comparand)



      	if (sentinel_counter<ending_point){                      // if this individual is far enough to the right to be compared
         	if(population[j]==-1){
               if (shared_flag==0)unshared++;
               shared_flag=0;
            	j++;
            	sentinel_counter++;

				}
         	if(population[j]==-2)break;
            if(population[i]==population[j])shared_flag=1;
       	}



       	if (sentinel_counter>=ending_point){
            if(population[j]==-2)break;

            if(population[j]==-1){
            	shared_flag=0;
            }
         }


      }
      i++;
      if(population[i]==-1){
      	i++;
		   individuals++;
         ending_point++;
      }
   }



  // printf("\n\rCount is %f",(double)unshared/(0.5*((individuals*individuals)-individuals)));getchar();


	 return (double)unshared/(0.5*((individuals*individuals)-individuals));

}


















